/*
 * IBM Confidential
 * OCO Source Materials
 * 5724-U98, 5655-V48, 5724-U50, 5655-V46
 *  Copyright IBM Corp. 2008, 2009
 * The source code for this program is not published or other -
 * wise divested of its trade secrets, irrespective of what has
 * been deposited with the U.S. Copyright Office.
 */
header {
	package com.ibm.nex.parser.oim.zos;
}

/*
 * OEF z/OS Parser
 */
{ @SuppressWarnings("all") }
class ZosParser extends Parser;

options {
	buildAST = true;
	k = 4;
}

tokens {
	// These tokens are defined because ANTLR 2.76 does not automatically generate constants for literals that contain a number. 
	LITERAL_Q1="Q1";
	LITERAL_Q2="Q2";
	LITERAL_101="101";
	LITERAL_NOT_SPECIFIED="NOT-SPECIFIED";
}

{		
	/**
	 * The characters that are invalid for an optim name.
	 * NOTE: This string comes from Optim Distributed code - rtrdba.cpp, gInvalidChars variable.
	 * In addition, the '%' character is invalid for names on import and export.
	 * Whitespace is not explicitly listed because tokens have been definied for whitespace, and such tokens
	 * should be skipped automatically.
	 */
	protected static final String INVALID_OPTIM_NAME_CHARACTERS = "%~`!^&*()+=|{}[]:;\"\'<>,.?/\\";	

	/**
	 * Determines if the token is a valid optim identifier.
	 */
	protected boolean isValidOptimIndentifierToken(Token token){
		boolean isValidToken = true;		
		String tokenText = token.getText();
		// Valid values are letters, '@', '#', '$', and '_'.  Digits are valid except as the first character.
		for(int charIndex=0; charIndex<tokenText.length(); charIndex++){
			char tokenCharacter = tokenText.charAt(charIndex);
			if (!Character.isLetter(tokenCharacter)){
				if ((charIndex > 0) && !Character.isDigit(tokenCharacter)){
					switch(tokenCharacter){
						case '@':
						case '#':
						case '$':
						case '_':
							break;
						default:
							return false;
					}
				} 
			}
		}
		return true;
	}

	/**
	 * Determines if the token is a valid optim name.
	 */
	protected boolean isValidOptimNameToken(Token token){
		String tokenText = token.getText();
		for(int charIndex=0; charIndex<tokenText.length(); charIndex++){
			char tokenCharacter = tokenText.charAt(charIndex);
			// If the character is invalid for an optim name, return false.
			if (INVALID_OPTIM_NAME_CHARACTERS.indexOf(tokenCharacter) != -1){
				return false;
			}
		}		
		return true;		
	}
	
	/**
	 * Determines if the token is a valid unsigned integer.
	 */
	protected boolean isValidUnsignedIntegerToken(Token token){
		return isValidUnsignedInteger(token.getText());
	}	

	/**
	 * Determines if the string is a valid unsigned integer.
	 */
	protected boolean isValidUnsignedInteger(String string){
		for(int charIndex=0; charIndex<string.length(); charIndex++){
			if (!Character.isDigit(string.charAt(charIndex))){
				return false;
			}
		}
		return true;
	}	
	
	/**
	 * Determines if the token is a valid variable delimiter.
	 */
	protected boolean isValidVariableDelimiter(Token token, boolean allowZero){
		boolean isValidToken = false;
		String tokenText = token.getText();
		if (tokenText.length() == 1){
			char tokenCharacter = tokenText.charAt(0);
			switch(tokenCharacter){
				case '@':
				case '#':
				case '$':
					isValidToken = true;
					break;
				case '0':
					isValidToken = allowZero;
					break;
			}
		}
		return isValidToken; 
	}
	
}

/*
 * CREATE
 */
create
	: ("CREATE"
		( accessDefinition
		| columnMap 
		| primaryKey
		| relationship
		| tableMap
		| archiveCollection
		| compareDefinition
		| environmentDefinition
		| legacyTable
		| retrievalDefinition		
  		)
	)+
	;

/**
 * AD
 */
accessDefinition
	: "AD"
	  (QUOTEDSTRING | accessDefinitionName)
	  (description)?
	  (security)?
	  (owner)?
	  lastModifiedTime
	  "DEFCID" creatorID
	  "START" qualifiedTableName
	  "ADDTBLS" yesNoChoice 
      "MODCRIT" yesNoChoice
      "BEGINDISP" accessDefinitionDisplayOption
      "ADCHGS" accessDefinitionChangePermission
      "USENEW" yesNoChoice
      ("APPLYSELFREF" yesNoChoice)?
      ("EXPIREVALUE" expirationValue)?
      ("ROWLIST" dataSetName)?
      (group)?
      (accessDefinitionTable)+
      (accessDefinitionRelationship)*
      (substitutionVariable)*
      SEMI
	;

/**
 * Access Definition table.
 * NOTE: Legacy parameters have been deliberately excluded for now.
 */
accessDefinitionTable
	: "TABLE" LEFTPAREN qualifiedTableName "ACCESS" tableAccessString "REF" yesNoChoice
	  ("ARCIXTAB" yesNoChoice)?
	  ("ARCDAA" yesNoChoice)?
	  ("ARCOP" andOrChoice)?
	  ("EXTRFREQ" unsignedInteger)? 
	  ("EXTRLIMIT" unsignedInteger)?
	  predicateOperator
	  "COLFLAG" yesNoChoice
	  (sqlClause)?
	  ("CORRELNAME" correlationName)?
	  ("ARC_ACTION" LEFTPAREN (archiveAction)+ RIGHTPAREN)?
	  (accessDefinitionColumn)*
	  RIGHTPAREN	
	;

accessDefinitionColumn
	: "COLUMN" LEFTPAREN 
	  columnName
	  "DISP" yesNoChoice "HEADING" columnHeading
	  ("SORT" sortSettings)?
	  "ACCESS" columnAccessString
	  ("ARCIXCOL" archiveIndexColumnType)?
	  ("PRED" delimitedStrings)?
	  ("ARCFMT" QUOTEDSTRING)?
	  ("ARCDATE" QUOTEDSTRING)?
	  ("ARCPIVOT" unsignedInteger)?
	  ("ARCYRS" unsignedInteger)?
	  ("ARCMOS" unsignedInteger)?
	  ("ARCWKS" unsignedInteger)?
	  ("ARCDAYS" unsignedInteger)?
      RIGHTPAREN
	;

accessDefinitionRelationship
	: "REL" LEFTPAREN relationshipName
	  "STATUS" relationshipStatusType
	  "Q1" yesNoChoice "Q2" (extraHyphen)? yesNoChoiceWithHyphen
	  "TYPE" relationshipType 
	  "DUPRELS" yesNoChoice
	  "CHILDLIMIT" (unsignedInteger | "NOCHILDREN" | "UNLIMITED")
	  ("PAR_ACCESS" accessMethod)?
	  ("PAR_KEYLIMIT" unsignedInteger)?
	  ("CHI_ACCESS" accessMethod)?
	  ("CHI_KEYLIMIT" unsignedInteger)?
	  "PAR" qualifiedTableName 
	  "CHI" qualifiedTableName
	  RIGHTPAREN
	;

/**
 * A flaw in zOS sometimes causes a hyphen to be added when none is needed.
 */
extraHyphen
	: GENERICOPTIMTOKEN
	;

/**
 * A flaw in zOS sometimes causes a hyphen to be added when none is needed.
 */
yesNoChoiceWithHyphen
	: yesNoChoice |
	  "Y-" |
	  "N-"
	;

/**
 * A named archive collection.
 * TODO: Modify this to read the specific grammar. 
 */
archiveCollection
	: "AC" (~SEMI)+ SEMI
	;

/**
 * A named column map
 */
columnMap
	: "CM" fullyQualifiedObjectDefinitionName
	  (description)?
	  (security)?
	  (owner)?
	  lastModifiedTime
	  ("EXTDSN" QUOTEDSTRING)?
	  "SRC" fullyQualifiedTableName
	  "DEST" fullyQualifiedTableName
	  "VAL" onOffChoice
	  columnMapEntries
	  SEMI
	  ;

columnMapEntries
	: LEFTPAREN columnMapExpression (COMMA columnMapExpression)* RIGHTPAREN	
	;

/**
 * A column map local to another object.
 */
columnMapLocal
	: columnMapEntries
	;


columnMapExpression
	: "SRCTYPE" sourceExpressionCode (delimitedStrings | columnMapExpressionPredicate) EQUALS columnNameOrQuotedString
	;

sourceExpressionCode
	: "A"
	| "B"
	| "C"
	| "D"
	| "E"
	| "F"
	| "H"
	| "I"
	| "J"
	| "K"
	| "L"
	| "M"
	| "N"
	| "P"
	| "T"
	| "U"
	| "X"
	| "Y"
	;

/**
 * A column expression predicate.
 * name and floatNumber can conflict at k==2.  This could happen for a single integer without a plus sign.
 * For example, "3" could be either a name or a floatNumber.  
 * In such a case, name takes precedence because it comes first.
 */
columnMapExpressionPredicateOLD
	: ((literalPrefix)? QUOTEDSTRING) | columnMapFunction | columnName | floatNumber
	;

columnMapExpressionPredicate
	: (~EQUALS)+
	;

columnMapFunction
	: columnMapFunctionName LEFTPAREN columnMapFunctionParameter (COMMA columnMapFunctionParameter)* RIGHTPAREN
	;

columnMapFunctionName
	: genericOptimToken
	;

columnMapFunctionParameter
	: QUOTEDSTRING | columnName | floatNumber
	;

/**
 * A named compare definition.
 * TODO: Modify this to read the specific grammar. 
 */
compareDefinition
	: "CD" (~SEMI)+ SEMI
	;

/**
 * A named environment definition.
 * TODO: Modify this to read the specific grammar. 
 */
environmentDefinition
	: "ED" (~SEMI)+ SEMI
	;

/**
 * A named legacy table.
 * TODO: Modify this to read the specific grammar. 
 */
legacyTable
	: "LT" (~SEMI)+ SEMI
	;

/**
 * A named retrieval definition.
 * TODO: Modify this to read the specific grammar. 
 */
retrievalDefinition
	: "RD" (~SEMI)+ SEMI
	;

tableMap
	: "TM" fullyQualifiedObjectDefinitionName
	  (description)?
	  (security)?
	  (owner)?
	  lastModifiedTime
	  "SRCCID" creatorID
	  "DESTCID" creatorID
	  ("COLMAPID" creatorID)?
	  (("SRCEXT" dataSetName) | ("SRCAD" accessDefinitionName))
	  (("DESTEXT" dataSetName) | ("DESTAD" accessDefinitionName))?
	  "VALRULES" tableMapValidationRules
	  ("SRCTYPE" tableMapSourceType)?
	  ("DESTTYPE" tableMapDestinationType)?
	  LEFTPAREN tableMapEntry (COMMA tableMapEntry)* RIGHTPAREN
	  SEMI
	;

tableMapEntry
	: qualifiedTableName EQUALS ("NOT-SPECIFIED" | qualifiedTableName)
	  (ownedColumnMap)?
	  ("ARC_ACTION" LEFTPAREN (archiveAction)+ RIGHTPAREN)?
	  ("PROCESSMODE" tableMapEntryProcessMode)?
	  ("DELETEFLAG" yesNoChoice)?
	  ("KEY_LIMIT" unsignedInteger)?
	;

tableMapEntryProcessMode
	: "I"
	| "U"
	| "B"
	;

ownedColumnMap
	: (("CM" qualifiedObjectDefinitionName) | ("LOCALCM" columnMapLocal))
	;

tableMapValidationRules
	: "M"
	| "C"
	;

tableMapSourceType
	: tableMapDataType
	;

tableMapDestinationType
	: tableMapDataType
	;

tableMapDataType
	: "X"
	| "A"
	| "T"
	;

literalPrefix
	: "X"
	| "G"
	| "GX"
	| "BX"
	;

onOffChoice
	: "ON" 
	| "OFF"
	;

substitutionVariable
	: "SUBVAR" LEFTPAREN
	  COLON variableName 
      "VALUELEN" "101"
      ("PROMPT" DELIMITEDSTRING)?
	  ("VALUE" DELIMITEDSTRING)?
	  RIGHTPAREN
	;

variableName
	: optimIdentifier
	;

/**
 * Access method - (K)ey lookup or Table (S)can.
 */
accessMethod
	: "K"
	| "S"
	;

/**
 * The relationship type
 */
relationshipType
	: "DB2"
	| "PST"
	| "OPT"
	| "IMS"
	;

/**
 * Relationship status type.  (NEW) to the relationship list, (SEL)ected, (UNSEL)ected, or (UNK)nown.
 */
relationshipStatusType
	: "NEW"
	| "SEL"
	| "UNSEL"
	| "UNK"
	;


/**
 * Archive Index Column Type - (D)ense index, (N)o index, or (S)parse index.
 */
archiveIndexColumnType
	: "D"
	| "N"
	| "S"
	;

/**
 * Sort settings - a token that combines the sort position with (A)scending or (D)escending order. 
 */
sortSettings {
		Token token = LT(1);
		String tokenText = token.getText();
		// Last character must be A or D.
		char sortDirection = tokenText.charAt(tokenText.length() - 1);
		switch(sortDirection){
			case 'A':
			case 'D':
				// Valid
				break;
			default:
				throw new RecognitionException("Invalid value for sortSettings rule.", getFilename(), token.getLine(), token.getColumn());
		}
		if(!isValidUnsignedInteger(tokenText.substring(0, tokenText.length() - 1))){
			throw new RecognitionException("Invalid value for sortSettings rule.", getFilename(), token.getLine(), token.getColumn());
		}
	}
	: genericOptimToken
	;

columnAccessString
	: "U" 
	| "S"
	;

/**
 * The column heading - a two character field where 
 * the first column represents the header text, and the second column 
 * describes the text position.  
 * Since the two characters will be parsed as a single token, 
 * we list all of the combinations.
 */
columnHeading
	: "NL"
	| "NR"
	| "NC"
	| "LL"
	| "LR"
	| "LC"
	;

archiveAction
	: LEFTPAREN
	archiveActionType shareStatus sqlStatement 
	RIGHTPAREN
	;

correlationName
	: optimName
	;

archiveActionType
	: "BEFORE_EXT_1ST_ROW"
	| "BEFORE_EXT_ROW"
	| "AFTER_EXT_ROW"
	| "AFTER_EXT_LAST_ROW"
	| "BEFORE_DEL_1ST_ROW"
	| "BEFORE_DEL_ROW"
	| "AFTER_DEL_ROW"
	| "AFTER_DEL_LAST_ROW"
	| "BEFORE_REST_1ST_ROW"
	| "BEFORE_REST_ROW"
	| "AFTER_REST_ROW"
	| "AFTER_REST_LAST_ROW"
	;

shareStatus
	: "NOT_SHARED"
	| "OWNER"
	;

sqlClause
	: "SQL" sqlStatement
	;

sqlStatement
	: delimitedStrings
	;

delimitedStrings
	: (DELIMITEDSTRING)+
	;

predicateOperator
	: "PREDOP" andOrChoice
	;

andOrChoice
	: "A"
	| "O"
	;

/**
 * Describes table access.  Must contain one or more of the following characters:
 * 		S, U, I, D
 * and only in that order.
 */
tableAccessString 
	: "S" |
	  "U" |
	  "I" |
	  "D" |
	  "SU" |
	  "SI" |
	  "SD" |
	  "UI" |
	  "UD" |
	  "ID" |
	  "SUI" |
	  "SUD" |
	  "SID" |
	  "UID" |
	  "SUID"
	;

/**
 * An access definition name.  
 */
accessDefinitionName
	: optimName DOT optimName DOT optimName
	;

/**
 * Display option - The initial display when an Access Defintion is used to edit or browse.
 * (D)ata
 * Selection criteria prompt for the (S)tart table
 * Selection criteria prompt for each (A)ccessed table, including the Start table.
 */
accessDefinitionDisplayOption
	: "D"
	| "S"
	| "A"
	;

/**
 * The access definition change permission setting.
 * Need to confirm correct values.
 */
accessDefinitionChangePermission
	: "T"
	| "P"
	;

/**
 * Expiration value - one of the following:
 * - a date in yyyy-mm-dd format.
 * - A 5 digit number representing number of days until expiration.
 * Both options can be represented by a single genericOptimToken.
 */
expirationValue
	: genericOptimToken
	;

/**
 * Data Set name
 */
dataSetName
	: QUOTEDSTRING | genericOptimToken (DOT genericOptimToken)*
	;

group
	: "GRPCOL" columnName "GRPROWS" unsignedInteger "GRPVALS" unsignedInteger
	;

primaryKey
	: "PK" primaryKeyName
	  (description)?
	  (security)?
	  (owner)?
	  lastModifiedTime
	  "COLS" LEFTPAREN columnName (COMMA columnName)* RIGHTPAREN
	  SEMI
	;

/**
 * A primary key name.
 */
primaryKeyName
	: creatorID DOT tableName
	;

relationship
	: "REL" relationshipName
	  (description)?
	  (security)?
	  (owner)?
	  lastModifiedTime
	  "CHI" fullyQualifiedTableName
	  "PAR" fullyQualifiedTableName
	  LEFTPAREN relationshipPredicateAssignment (COMMA relationshipPredicateAssignment)* RIGHTPAREN
	  SEMI
	;

relationshipPredicateAssignment
	: relationshipPredicate EQUALS relationshipPredicate
	;

/**
 * A relationship predicate.
 * name and floatNumber can conflict at k==2.  This could happen for a single integer without a plus sign.
 * For example, "3" could be either a name or a floatNumber.  
 * In such a case, name takes precedence because it comes first.
 */
relationshipPredicate	
	: "NULL" | 
	  "TRUE" | 
      "FALSE" |        
      columnNameOrQuotedString | 
      floatNumber
	;

fullyQualifiedObjectDefinitionName
	: creatorID DOT objectName
	;

qualifiedObjectDefinitionName
	: (creatorID DOT)? objectName
	;

objectName
	: optimIdentifier
	;

/**
 * Because of the way column name is defined, quoted strings are already accepted.
 */
columnNameOrQuotedString
	: columnName
	;

relationshipName
	: databaseObjectName
	;

/**
 * A fully qualified table name.
 * Similar to qualifiedTableName, except all parts are required.
 */
fullyQualifiedTableName
	: creatorID DOT tableName
	;

/**
 * A qualified table name - can have creatorID DOT tableName, or just tableName
 * NOTE: Due to ANTLR issues, this cannot be specified as:
 *	(creatorID DOT)? tableName
 * The parser code generated to look for the first clause is incorrect, and there is no easy solution.
 * Since both parts are really just quotableLongNames, the workaround is to make the first part required,
 * with the DOT and second part optional.
 */
qualifiedTableName
	: quotableLongName (DOT quotableLongName)?
	;

creatorID
	: quotableLongName
	;

tableName
	: databaseObjectName
	;

description
	: "DESCRIPTION" DELIMITEDSTRING
	;

security
	: "SECURITY" securityType
	;

securityType
	: "PUBLIC"
	| "READONLY"
	| "PRIVATE"
	;

// Fix references to genericOptimToken
owner
	: "OWNER" genericOptimToken
	;

// Fix references to genericOptimToken
lastModifiedTime
	: "MODIFIED" genericOptimToken DOT unsignedInteger DOT unsignedInteger
	;

/**
 * A database column name
 */
columnName
	: databaseObjectName
	;

/**
 * A database object name.  Used for things that can exist in a database, like tables, columns, views, procedures, packages, etc.
 */
databaseObjectName
	: quotableLongName
	;

/**
 * Text that can either span multiple lines or be enclosed in quotes.
 * Items enclosed in quotes may also span multiple lines. 
 */
quotableLongName
	: QUOTEDSTRING | longName
	;

/**
 * Text that can span multiple lines
 * Text is spanned as multiple tokens separated by newlines.
 */
longName
	: (optimName)+
	;

floatNumber {
		/* 
			This section insures that a plus is not followed by a minus.
		*/
		if (LA(1) == PLUS && LA(2) == GENERICOPTIMTOKEN) {
			Token token = LT(2);
			if (token.getText().charAt(0) == '-') {
				throw new RecognitionException("Invalid value for floatNumber rule.", getFilename(), token.getLine(), token.getColumn());
			}
		}		
	}
	: (	(PLUS)?
		( 
	    	minusOrInteger (DOT (unsignedInteger)?)? |
	     	DOT unsignedInteger
	   	) 
	  )
	;

/** 
 * A minus sign, unsigned integer, or a minus sign followed by an integer.
 * The action block checks that the characters are valid.
 */		
minusOrInteger {
		Token token = LT(1);
		String tokenText = token.getText();
		boolean isValidToken = true;
		int charIndex = 0;
		if (tokenText.charAt(charIndex) == '-'){
			charIndex++;
		}
		if (!isValidUnsignedInteger(tokenText.substring(charIndex))){
			throw new RecognitionException("Invalid value for minusOrInteger rule.", getFilename(), token.getLine(), token.getColumn());
		}		
	}
	: GENERICOPTIMTOKEN
	;

/** 
 * An unsigned integer.
 * Unsigned integers contain a subset of characters that can be in a GENERICOPTIMTOKEN.
 * The action block checks that the characters are valid.
 */		
unsignedInteger {
		Token token = LT(1);
		if (!isValidUnsignedIntegerToken(token)){
			throw new RecognitionException("Invalid value for unsignedInteger rule.", getFilename(), token.getLine(), token.getColumn());			
		}
	}
	: genericOptimToken
	;

/**
 * An identifier recognized by Optim.
 * This rule is intended to handle Optim Distributed methods RtRDBA::IsPSTNameValid 
 * and RtRDBA::ValidPSTIdentifier.
 * NOTE: It is recommended that simple rules refer to this rule as a way of clarifying how the identifier is being used.
 * For example, dbAliasName is an optimIdentifier.
 * Please do NOT use this rule directly from a larger definition, such as accessDefinition.
 * Doing so obscures the actual purpose of the rule.  
 */
optimIdentifier 
	: {
		Token token = LT(1);
		if (!isValidOptimIndentifierToken(token)){
			throw new RecognitionException("Invalid value for optimIdentifier rule.", getFilename(), token.getLine(), token.getColumn());			
		}
	} genericOptimToken
	;

optimName
	: genericOptimToken
	;

/**
 * A name recognized by Optim.
 * This rule is intended to handle Optim Distributed method RtRDBA::IsNameValid.
 * NOTE: It is recommended that simple rules refer to this rule as a way of clarifying how the name is being used.
 * For example, creatorID is an optimName.
 * Please do NOT use this rule directly from a larger definition, such as accessDefinition.
 * Doing so obscures the actual purpose of the rule.  
 */
 genericOptimToken
	: {
		/* 
			This section handles cases where a keyword is really an GENERICOPTIMTOKEN.
			
			Most grammars to not allow a name to be a keyword.
			For example, in Java, you cannot have an int named "int".
			However, Optim Export Files *do* allow a name to be a keyword.  
			For example, an export file can have a statement to create an Access Definition 
			that begins with "CREATE AD AD.AD" where "AD" is a keyword, 
			but "AD.AD" is the name of the Access Definition.
			
			To handle this condition, the following code checks to see if the token is 
			the expected type (in this case, an GENERICOPTIMTOKEN).  If it is not, then we check to see 
			if the token is a keyword *that has the properties* of an GENERICOPTIMTOKEN.  
			If so, we force the token to behave like an GENERICOPTIMTOKEN.
			
			All keywords in OEF are simple strings of letters (and sometimes an underscore).
			Therefore, to determine if the keyword is really an identifier, the code just checks 
			that the keyword starts with a letter.  
			
			This fix has a potential side-effect.  If the creator of an OEF forgets a name, 
			the parser will see the next intended keyword, force it to be a different type, 
			and accept it. The net result is that a generated error will pertain to the next part 
			of the OEF file, probably complaining about an expected keyword 
			(namely, the one whose type was just changed).

			NOTE: THIS IS NOT A CATCH-ALL SOLUTION.
			In cases where alternatives are presented (e.g., "rule1 | rule2"), ANTLR generates code that
			looks ahead to determine which rule to follow.  The lookahead is based on tokens,
			and therefore the rule code is not executed before the token is analyzed.
			I've tested with making the token in this rule match any token using the "." operator.
			The problem there is that too many tokens are allowed - in a rule "(genericOptimToken)+ RIGHTPAREN",
			the RIGHTPAREN is treated as a genericOptimToken, and then fails the validation.
			
			Since the issue is literal tokens listed above, one possible solution is to use 
			a rule that lists *all* literal tokens.  Then, this rule would be 
			"GENERICOPTIMTOKEN | anyLiteral".
			
		*/						
		if (LA(1) != GENERICOPTIMTOKEN) {
			Token token = LT(1);
			if (isValidOptimNameToken(token)) {
				token.setType(GENERICOPTIMTOKEN);
			} else {
				throw new RecognitionException("Invalid value for genericOptimToken rule.", getFilename(), token.getLine(), token.getColumn());				
			}
		}		
	} GENERICOPTIMTOKEN
	;

yesNoChoice
    : "Y"
    | "N"
    ;

/*
 * OEF z/OS Lexer
 */
{ @SuppressWarnings("all") }
class ZosLexer extends Lexer;

options {
	k = 5;
	charVocabulary = '\u0000'..'\uFFFE';
}

DELIMITEDSTRING
	: "//"
	(
		options {
			greedy = false;  // Make it exit upon "//"
			generateAmbigWarnings = false; // Shut off newline errors
		}
		:	'\r' '\n'	{newline();}
		|	'\r'		{newline();}
		|	'\n'		{newline();}
		|	~('\n'|'\r')
		)*
	"//"
	;
 
/**
 * Used for identifiers such as definition IDs and names.
 * A string of one or more characters where special characters are excluded.
 * The order of the characters after the '%' comes from rtrdba.cpp gInvalidChars.
 * This token is also used for integers.
 *
 * NOTE: We cannot use a separate token for integers.
 */
GENERICOPTIMTOKEN
	: (
		~(
			' ' | 
			'\t'| 
			'\u001a' |
			'\n' |
			'\r' |
			'%' |
			'~' |
			'`' |
			'!' |
			'^' |
			'&' |
			'*' |
			'(' |
			')' |
			'+' |
			'=' |
			'|' |
			'{' |
			'}' |
			'[' |
			']' |
			':' |
			';' |
			'\"' |
			'\'' |
			'<' |
			'>' |
			',' |
			'.' |
			'?' |
			'/' |
			'\\'			
	 	)
	)+
	;

QUOTEDSTRING
	: (((DOUBLEQUOTE)(~'"')*(DOUBLEQUOTE)) | ((SINGLEQUOTE)(~'\'')*(SINGLEQUOTE)))
	;

AMPERSAND
	: '&'
	;

BACKSLASH
	: '\\'
	;

BANG
	: '!'
	;

CARET
	: '^'
	;

COLON
	: ':'
	;

COMMA
	: ','
	;

DOT
	: '.'
	;

DOUBLEQUOTE
	: '"'
	;

EQUALS
	: '='
	;

GREATERTHAN
	: '>'
	;

LEFTPAREN
	: '('
	;

LESSTHAN
	: '<' 
	;

PERCENT
	: '%'
	;

DOUBLEPIPE
	: PIPE PIPE
	;

PIPE
	: '|'
	;

PLUS
	: '+'
	;

QUESTION
	: '?'
	;

RIGHTPAREN
	: ')'
	;

SEMI
	: ';'
	;

STAR
	: '*'
	;

SINGLEQUOTE
	: '\''
	;

TILDE
	: '~'
	;

SLASH
	: '/'
	;

NEWLINE
	: ("\r\n" | '\n' | '\r') {$setType(Token.SKIP); newline();}
	;
WS
	: (' ' | '\t' | '\u001a') {$setType(Token.SKIP);}
	;

